@@ -13,7 +13,6 @@
 
 """Handling introspection data from the ramdisk."""
 
-import base64
 import copy
 import datetime
 import os
@@ -22,12 +21,14 @@ import eventlet
 import json
 
 from oslo_config import cfg
+from oslo_serialization import base64
 from oslo_utils import excutils
 
 from ironic_inspector.common.i18n import _, _LE, _LI, _LW
 from ironic_inspector.common import ironic as ir_utils
 from ironic_inspector.common import swift
 from ironic_inspector import firewall
+from ironic_inspector import introspection_state as istate
 from ironic_inspector import node_cache
 from ironic_inspector.plugins import base as plugins_base
 from ironic_inspector import rules
@@ -73,7 +74,7 @@ def _store_logs(introspection_data, node_info):
             os.makedirs(CONF.processing.ramdisk_logs_dir)
         with open(os.path.join(CONF.processing.ramdisk_logs_dir, file_name),
                   'wb') as fp:
-            fp.write(base64.b64decode(logs))
+            fp.write(base64.decode_as_bytes(logs))
     except EnvironmentError:
         LOG.exception(_LE('Could not store the ramdisk logs'),
                       data=introspection_data, node_info=node_info)
@@ -92,6 +93,11 @@ def _find_node_info(introspection_data, failures):
         if not_found_hook is None:
             failures.append(_('Look up error: %s') % exc)
             return
+
+        LOG.debug('Running node_not_found_hook %s',
+                  CONF.processing.node_not_found_hook,
+                  data=introspection_data)
+
         # NOTE(sambetts): If not_found_hook is not none it means that we were
         # unable to find the node in the node cache and there is a node not
         # found hook defined so we should try to send the introspection data
@@ -110,6 +116,8 @@ def _find_node_info(introspection_data, failures):
 def _run_pre_hooks(introspection_data, failures):
     hooks = plugins_base.processing_hooks_manager()
     for hook_ext in hooks:
+        LOG.debug('Running pre-processing hook %s', hook_ext.name,
+                  data=introspection_data)
         # NOTE(dtantsur): catch exceptions, so that we have changes to update
         # node introspection status after look up
         try:
@@ -227,7 +235,7 @@ def process(introspection_data):
             _store_logs(introspection_data, node_info)
 
     try:
-        result = _process_node(node, introspection_data, node_info)
+        result = _process_node(node_info, node, introspection_data)
     except utils.Error as exc:
         node_info.finished(error=str(exc))
         with excutils.save_and_reraise_exception():
@@ -251,15 +259,17 @@ def _run_post_hooks(node_info, introspection_data):
     hooks = plugins_base.processing_hooks_manager()
 
     for hook_ext in hooks:
+        LOG.debug('Running post-processing hook %s', hook_ext.name,
+                  node_info=node_info, data=introspection_data)
         hook_ext.obj.before_update(introspection_data, node_info)
 
 
-def _process_node(node, introspection_data, node_info):
+@node_cache.fsm_transition(istate.Events.process, reentrant=False)
+def _process_node(node_info, node, introspection_data):
     # NOTE(dtantsur): repeat the check in case something changed
     ir_utils.check_provision_state(node)
-
-    node_info.create_ports(introspection_data.get('macs') or ())
-
+    interfaces = introspection_data.get('interfaces')
+    node_info.create_ports(list(interfaces.values()))
     _run_post_hooks(node_info, introspection_data)
     _store_data(node_info, introspection_data)
 
@@ -275,18 +285,20 @@ def _process_node(node, introspection_data, node_info):
         new_username, new_password = (
             node_info.options.get('new_ipmi_credentials'))
         utils.executor().submit(_finish_set_ipmi_credentials,
-                                ironic, node, node_info, introspection_data,
+                                node_info, ironic, node, introspection_data,
                                 new_username, new_password)
         resp['ipmi_setup_credentials'] = True
         resp['ipmi_username'] = new_username
         resp['ipmi_password'] = new_password
     else:
-        utils.executor().submit(_finish, ironic, node_info, introspection_data)
+        utils.executor().submit(_finish, node_info, ironic, introspection_data,
+                                power_off=CONF.processing.power_off)
 
     return resp
 
 
-def _finish_set_ipmi_credentials(ironic, node, node_info, introspection_data,
+@node_cache.fsm_transition(istate.Events.finish)
+def _finish_set_ipmi_credentials(node_info, ironic, node, introspection_data,
                                  new_username, new_password):
     patch = [{'op': 'add', 'path': '/driver_info/ipmi_username',
               'value': new_username},
@@ -305,12 +317,12 @@ def _finish_set_ipmi_credentials(ironic, node, node_info, introspection_data,
             ironic.node.get_boot_device(node_info.uuid)
         except Exception as exc:
             LOG.info(_LI('Waiting for credentials update, attempt %(attempt)d '
-                         'current error is %(exc)s') %
+                         'current error is %(exc)s'),
                      {'attempt': attempt, 'exc': exc},
                      node_info=node_info, data=introspection_data)
             eventlet.greenthread.sleep(_CREDENTIALS_WAIT_PERIOD)
         else:
-            _finish(ironic, node_info, introspection_data)
+            _finish_common(node_info, ironic, introspection_data)
             return
 
     msg = (_('Failed to validate updated IPMI credentials for node '
@@ -319,7 +331,7 @@ def _finish_set_ipmi_credentials(ironic, node, node_info, introspection_data,
     raise utils.Error(msg, node_info=node_info, data=introspection_data)
 
 
-def _finish(ironic, node_info, introspection_data, power_off=True):
+def _finish_common(node_info, ironic, introspection_data, power_off=True):
     if power_off:
         LOG.debug('Forcing power off of node %s', node_info.uuid)
         try:
@@ -328,7 +340,7 @@ def _finish(ironic, node_info, introspection_data, power_off=True):
             if node_info.node().provision_state == 'enroll':
                 LOG.info(_LI("Failed to power off the node in"
                              "'enroll' state, ignoring; error was "
-                             "%s") % exc, node_info=node_info,
+                             "%s"), exc, node_info=node_info,
                          data=introspection_data)
             else:
                 msg = (_('Failed to power off node %(node)s, check '
@@ -346,6 +358,9 @@ def _finish(ironic, node_info, introspection_data, power_off=True):
              node_info=node_info, data=introspection_data)
 
 
+_finish = node_cache.fsm_transition(istate.Events.finish)(_finish_common)
+
+
 def reapply(node_ident):
     """Re-apply introspection steps.
 
@@ -375,37 +390,52 @@ def _reapply(node_info):
     # runs in background
     try:
         introspection_data = _get_unprocessed_data(node_info.uuid)
-    except Exception:
+    except Exception as exc:
         LOG.exception(_LE('Encountered exception while fetching '
                           'stored introspection data'),
                       node_info=node_info)
-        node_info.release_lock()
-        return
-
-    failures = []
-    _run_pre_hooks(introspection_data, failures)
-    if failures:
-        LOG.error(_LE('Pre-processing failures detected reapplying '
-                      'introspection on stored data:\n%s'),
-                  '\n'.join(failures), node_info=node_info)
-        node_info.finished(error='\n'.join(failures))
+        msg = (_('Unexpected exception %(exc_class)s while fetching '
+                 'unprocessed introspection data from Swift: %(error)s') %
+               {'exc_class': exc.__class__.__name__, 'error': exc})
+        node_info.finished(error=msg)
         return
 
     try:
         ironic = ir_utils.get_client()
-        node_info.create_ports(introspection_data.get('macs') or ())
-        _run_post_hooks(node_info, introspection_data)
-        _store_data(node_info, introspection_data)
-        node_info.invalidate_cache()
-        rules.apply(node_info, introspection_data)
-        _finish(ironic, node_info, introspection_data,
-                power_off=False)
     except Exception as exc:
-        LOG.exception(_LE('Encountered exception reapplying '
-                          'introspection on stored data'),
-                      node_info=node_info,
-                      data=introspection_data)
+        msg = _('Encountered an exception while getting the Ironic client: '
+                '%s') % exc
+        LOG.error(msg, node_info=node_info, data=introspection_data)
+        node_info.fsm_event(istate.Events.error)
+        node_info.finished(error=msg)
+        return
+
+    try:
+        _reapply_with_data(node_info, introspection_data)
+    except Exception as exc:
         node_info.finished(error=str(exc))
-    else:
-        LOG.info(_LI('Successfully reapplied introspection on stored '
-                     'data'), node_info=node_info, data=introspection_data)
+        return
+
+    _finish(node_info, ironic, introspection_data,
+            power_off=False)
+
+    LOG.info(_LI('Successfully reapplied introspection on stored '
+                 'data'), node_info=node_info, data=introspection_data)
+
+
+@node_cache.fsm_event_before(istate.Events.reapply)
+@node_cache.triggers_fsm_error_transition()
+def _reapply_with_data(node_info, introspection_data):
+    failures = []
+    _run_pre_hooks(introspection_data, failures)
+    if failures:
+        raise utils.Error(_('Pre-processing failures detected reapplying '
+                            'introspection on stored data:\n%s') %
+                          '\n'.join(failures), node_info=node_info)
+
+    interfaces = introspection_data.get('interfaces')
+    node_info.create_ports(list(interfaces.values()))
+    _run_post_hooks(node_info, introspection_data)
+    _store_data(node_info, introspection_data)
+    node_info.invalidate_cache()
+    rules.apply(node_info, introspection_data)
